#!/bin/bash
# This script will work with one squid or up to 4 squids on the same http port.
# The number of squids is determined by the existence of cache directories
# as follows.  The main path to the cache directories is determined by the
# cache_dir option in squid.conf.  To run multiple squids, create directories
# of the form
#   $cache_dir/squid$N
# where N goes from 0 to the number of squids minus 1.  cache_log directories
# of the same form will be used, and created if needed.  Note that the
# cache_log option in squid.conf is a file, not a directory, so the squid$N
# is up one level:
#   cache_log_file=`basename $cache_log`
#   cache_log_dir=`dirname $cache_log`/squid$N
#   cache_log=$cache_log_dir/$cache_log_file
# The access_log should be in the same directory as the cache_log (unless it
# is set to "none"), and the pid_filename also will be in similarly named
# directories (the same directories as the cache_log is a good choice but not
# required, and the directories will be created if needed).
# Each squid also listens separately on port 3127-$N for http requests
#  in addition to the shared port 3128, and each squid listens on
#  port 3401+$N for snmp queries.  Only the first squid listens on 3130
#  for icp.  (All of those starting numbers can be overridden in squid.conf,
#  they're just the defaults.)  Each squid after the first one uses the
#  first one as a parent, so they don't all contact the upstream servers
#  separately.

. /etc/init.d/functions

RETVAL=0

CONF_DIR=/etc/squid2
SBIN_DIR=/usr/sbin
AWSTATS_SCRIPT=/etc/awstats/run_awstats.sh
SQUID_COMPRESS_LOGS=${SQUID_COMPRESS_LOGS:-true}
DEFAULT_MAX_ACCESS_LOG=1G
if [ "$SQUID_COMPRESS_LOGS" = true ]; then
    DEFAULT_MAX_ACCESS_LOG=5G
fi
#size at which rotateiflarge will rotate access.log
#can be in bytes, or if ending 'M' in megabytes, or if ending 'G' in gigabytes
#use LARGE_ACCESS_LOG if set for backward compatibility
SQUID_MAX_ACCESS_LOG=${SQUID_MAX_ACCESS_LOG:-${LARGE_ACCESS_LOG:-$DEFAULT_MAX_ACCESS_LOG}}
case $SQUID_MAX_ACCESS_LOG in
    *M) let SQUID_MAX_ACCESS_LOG="${SQUID_MAX_ACCESS_LOG%M}*1000000";;
    *G) let SQUID_MAX_ACCESS_LOG="${SQUID_MAX_ACCESS_LOG%G}*1000000000";;
esac
#set to false to not clean cache on start
SQUID_CLEAN_CACHE_ON_START=${SQUID_CLEAN_CACHE_ON_START:-true}

CONF_FILE=$CONF_DIR/squid.conf

if [ `id -u` = 0 ]; then
  echo "ERROR: `basename $0` not allowed to be run as root" >&2
  exit 1
fi

if [ ! -d "$CONF_DIR" ] || [ ! -w "$CONF_DIR" ]; then
 echo "ERROR: $CONF_DIR does not exist or is not writable" >&2
 exit 1
fi

# see if squid.conf needs to be regenerated
if [ $CONF_FILE.frontierdefault -nt $CONF_FILE ] ||
	[ $CONF_DIR/customize.sh -nt $CONF_FILE ] ||
	[ $0 -nt $CONF_FILE ]; then
 if [ -f "$CONF_FILE" ] && [ -n "`find "$CONF_FILE" -perm -0200 2>/dev/null`" ]; then
  echo "ERROR: $CONF_FILE is writable and was about to be overwritten" >&2
  echo "  This may indicate that squid was upgraded from an older version" >&2
  echo "  that was not based on customize.sh." >&2
  echo "  Remove or rename it, or chmod it to be non-writable" >&2
  exit 1
 fi
 echo "Generating $CONF_FILE"
 rm -f $CONF_FILE.new
 if $CONF_DIR/customize.sh <$CONF_FILE.frontierdefault >$CONF_FILE.new; then
  rm -f $CONF_FILE.old
  if [ -f $CONF_FILE ]; then
   echo "Saving old configuration as $CONF_FILE.old"
   mv $CONF_FILE $CONF_FILE.old
  fi
  mv $CONF_FILE.new $CONF_FILE
  chmod 444 $CONF_FILE
 else
  exit $?
 fi
fi

# assume there are no spaces in these file paths
CACHE_DIR=`awk '$1 == "cache_dir" {x=$3} END{print x}' $CONF_FILE`
CACHE_LOG=`awk '$1 == "cache_log" {x=$2} END{print x}' $CONF_FILE`
ACCESS_LOG=`awk '$1 == "access_log" {x=$2} END{print x}' $CONF_FILE`
PID_FILENAME=`awk '$1 == "pid_filename" {x=$2} END{print x}' $CONF_FILE`

if [ -z "$CACHE_DIR" ]; then
 echo "ERROR: cache_dir not found in $CONF_FILE" >&2
 exit 1
fi
if [ "$CACHE_DIR" == "/" ]; then
 echo "ERROR: cache_dir not allowed to be '/' in $CONF_FILE" >&2
 exit 1
fi
if [ ! -d "$CACHE_DIR" ] || [ ! -w "$CACHE_DIR" ]; then
 echo "ERROR: cache_dir $CACHE_DIR does not exist or is not writable" >&2
 exit 1
fi
if [ -z "$CACHE_LOG" ]; then
 echo "ERROR: cache_log not found in $CONF_FILE" >&2
 exit 1
fi
CACHE_LOG_DIR=`dirname "$CACHE_LOG"|sed 's/daemon://'`
if [ ! -d "$CACHE_LOG_DIR" ] || [ ! -w "$CACHE_LOG_DIR" ]; then
 echo "ERROR: cache_log directory $CACHE_LOG_DIR does not exist or is not writable" >&2
 exit 1
fi
if [ -z "$ACCESS_LOG" ]; then
 echo "ERROR: access_log not found in $CONF_FILE" >&2
 exit 1
fi
if [ "$ACCESS_LOG" != "none" ]; then
 ACCESS_LOG_DIR=`dirname "$ACCESS_LOG"|sed 's/daemon://'`
 if [ "$ACCESS_LOG_DIR" != "$CACHE_LOG_DIR" ]; then
  echo "ERROR: access_log directory $ACCESS_LOG_DIR is not the same as the cache_log directory $CACHE_LOG_DIR" >&2
  exit 1
 fi
else
 ACCESS_LOG_DIR="$CACHE_LOG_DIR"
fi
if [ -z "$PID_FILENAME" ]; then
 echo "ERROR: pid_filename not found in $CONF_FILE" >&2
 exit 1
fi
PID_FILENAME_DIR=`dirname "$PID_FILENAME"`
if [ ! -d "$PID_FILENAME_DIR" ] || [ ! -w "$PID_FILENAME_DIR" ]; then
 echo "ERROR: pid_filename directory $PID_FILENAME_DIR does not exist or is not writable" >&2
 exit 1
fi

LOGFILE_ROTATE=`awk '$1 == "logfile_rotate" {x=$2} END{print x}' $CONF_FILE`
if [ -z "$LOGFILE_ROTATE" ]; then
 LOGFILE_ROTATE=10
fi

squid_dirs()
{
 # if $NUMSQUIDS is 1, echo the parameter, otherwise echo the parameter N
 #   times with /squid$N after the parameter, where N is from 0 to $NUMSQUIDS-1

 if [ $NUMSQUIDS = 1 ]; then
  echo $1
  return
 fi
 typeset N=0
 typeset SQUID_DIRS=""
 while [ $N -lt $NUMSQUIDS ]; do
  SQUID_DIRS="$SQUID_DIRS $1/squid$N"
  let N=$N+1
 done
 echo $SQUID_DIRS
}

#see how many squid cache directories exist (up to 16)
NUMSQUIDS=0
for D in `NUMSQUIDS=16 squid_dirs $CACHE_DIR`; do
 if [ ! -d $D ]; then
  break
 fi
 let NUMSQUIDS=$NUMSQUIDS+1
done

SQUIDEXE=$SBIN_DIR/squid2
if [ $NUMSQUIDS = 0 ]; then
 NUMSQUIDS=1
 SQUID=$SQUIDEXE
 PLURAL=""
else
 PLURAL="s"
 if [ -z "$SETSQUIDAFFINITY" ]; then
  # When there's more than one squid (and no more cores than squids), default
  #  to set the affinity of each squid to one core for best performance.
  # Measurements showed a ~15% throughput boost per core when cpu-limited.
  let NUMCORES="$(grep ^processor /proc/cpuinfo|tail -1|sed 's/.*: //') + 1"
  if [ "$NUMCORES" -ge "$NUMSQUIDS" ]; then
   SETSQUIDAFFINITY=true
  fi
 fi

 #create corresponding $CACHE_LOG_DIR and $PID_FILENAME_DIR directories
 #  if needed
 for D in $(squid_dirs $CACHE_LOG_DIR) $(squid_dirs $PID_FILENAME_DIR); do
  if [ ! -d $D ]; then
   mkdir $D
  fi
 done

 # create the squid.conf files for each squid
 HTTP_PORT=`awk '$1 == "http_port" {print $2;exit}' $CONF_FILE`
 HTTP_PORT_EXTRA="`awk '$1 == "http_port" {$1="";$2="";print;exit}' $CONF_FILE`"
 SNMP_PORT=`awk '$1 == "snmp_port" {x=$2} END{print x}' $CONF_FILE`
 ICP_PORT=`awk '$1 == "icp_port" {x=$2} END{print x}' $CONF_FILE`
 #visible_hostname is the name that shows up in X-Cache and Via headers
 VISIBLE_HOSTNAME=`awk '$1 == "visible_hostname" {x=$2} END{print x}' $CONF_FILE`
 HOSTNAME=`hostname`

 if [ -z "$SQUID_MULTI_PEERING" ]; then
  if [ -z "`awk '$1 == "cache_peer" && $3 == "parent" {print;exit}' $CONF_FILE`" ]; then
   # enable all but the first squid to be parent of the others because user
   #  did not set $SQUID_MULTI_PEERING and there is no cache_peer parent
   #  defined
   SQUID_MULTI_PEERING=true
  fi
 fi

 SQUID="$(dirname $(dirname $0))/sbin/multisquid2 $SQUIDEXE $CONF_DIR $NUMSQUIDS ${HTTP_PORT:-3128}"

 #
 # create separate .squid-$N.conf for each squid process when needed
 #
 N=0
 while [ $N -lt $NUMSQUIDS ]; do
  NEWCONF=$CONF_DIR/.squid-$N.conf
  if [ ! $CONF_DIR/squid.conf -nt $NEWCONF ]; then
   # only create the per-squid config file if squid.conf is newer
   let N=$N+1
   continue
  fi
  rm -f $NEWCONF
  SEDCMDS="-e \"s,$CACHE_DIR,$CACHE_DIR/squid$N,\""
  if [ "$CACHE_LOG_DIR" != "$CACHE_DIR" ]; then
   SEDCMDS="$SEDCMDS -e \"s,$CACHE_LOG_DIR/,$CACHE_LOG_DIR/squid$N/,\""
  fi
  if [ "$PID_FILENAME_DIR" != "$CACHE_DIR" ] && [ "$PID_FILENAME_DIR" != "$CACHE_LOG_DIR" ]; then
   SEDCMDS="$SEDCMDS -e \"s,$PID_FILENAME_DIR/,$PID_FILENAME_DIR/squid$N/,\""
  fi
  eval sed $SEDCMDS $CONF_DIR/squid.conf | awk '
   BEGIN{print "# DO NOT EDIT -- Automatically generated by '$0'";print}
   /^[ \t#]*http_port[ \t]/ {
    if (!got_http_port) {
     #add additional http_port after first one specified, or after first
     #   commented http_port if none specified
     portnum=0
     if (substr($0,1,1) != "#") portnum=$2
     else if ("'$HTTP_PORT'" == "") {portnum=3128; print}
     if (portnum != 0) {
      print "#first port overridden by command line -I"
      print "http_port", portnum, "'"$HTTP_PORT_EXTRA"'"
      # have each squid also listen on their own port, the first one to
      # be parent of others and the others for forced cache reloads
      print "http_port", portnum - ('$N' + 1), "'"$HTTP_PORT_EXTRA"'"
      got_http_port=1
      next
     }
    }
    else if ((substr($0,1,1) != "#") && ("'$N'" != 0)) {
     # only first squid listens on other specified http ports, skip this one
     next
    }
   }
   /^[ \t#]*snmp_port[ \t]/ {
    #replace snmp_port with separate number for each squid
    portnum=0
    if (substr($0,1,1) != "#") portnum=$2
    else if ("'$SNMP_PORT'" == "") {portnum=3401; print}
    if (portnum != 0) {
     portnum=portnum + '$N'
     print "snmp_port", portnum
     next
    }
   }
   /^[ \t#]*icp_port[ \t]/ {
    #only first squid listens on icp port
    portnum=0
    if (substr($0,1,1) != "#") portnum=$2
    else if ("'$ICP_PORT'" == "") {portnum=3130; print}
    if (portnum != 0) {
     if ("'$N'" != 0) portnum=0  # disables the icp port
     print "icp_port", portnum
     next
    }
   }
   /^[ \t#]*cache_peer[ \t]/ {
    #any squid after first one go only to first one as a cache_peer parent
    #  unless $SQUID_MULTI_PEERING is not set to true
    if (("'$N'" != 0) && ("'"$SQUID_MULTI_PEERING"'" == "true")) {
     if (substr($0,1,1) == "#") {
      if (!got_cache_peer) {
       #insert the line after the first commented-out one
       print
       portnum='${HTTP_PORT:-3128}' - 1
       print "cache_peer localhost parent", portnum, "0 no-query"
       got_cache_peer=1
       next
      }
     }
     else {
      #skip any other uncommented cache_peer options on non-first squids
      next
     }
    }
   }
   /^ *update_headers[ \t]/ {
    if (("'$N'" == 0) && ("'$NUMSQUIDS'" > 0)) {
     # this has to be on for squids that feed other squids
     print "update_headers on"
    }
    else {
     # take default
     print
    }
    next
   }
   $1 == "visible_hostname" || /TAG: visible_hostname/ {
    hostname=""
    if (substr($0,1,1) != "#") hostname=$2
    else if ("'$VISIBLE_HOSTNAME'" == "") {hostname="'$HOSTNAME'"; print}
    if (hostname != "") {
     print "visible_hostname",hostname "/'$N'"
    }
   }
   {print}
   ' >$NEWCONF
  chmod a-w $NEWCONF
  let N=$N+1
 done
fi

isrunning()
{
 # return 0 if all running, else 1
 for D in `squid_dirs $PID_FILENAME_DIR`; do
  typeset MY_PID_FILENAME="$D/$(basename $PID_FILENAME)"
  if [ ! -f "$MY_PID_FILENAME" ]; then
   return 1
  fi
  typeset PID="`cat <$MY_PID_FILENAME 2>/dev/null`"
  if [ ! -n "$PID" ]; then
   rm -f "$MY_PID_FILENAME"
   return 1
  fi
  # Need to make sure that $PID actually refers to squid and not some
  #   other program after a system failure & restart.
  # Can't use pidof, because nightly prelink confuses it when $SBIN_DIR
  #   is a system directory.
  # Can't use ls -iL on /proc/$PID/exe, because that shows the inode
  #   of the executable before prelink messed with it, so instead extract
  #   the name of the symlink target and do -iL on that.
  typeset PIDEXE="`ls -l /proc/$PID/exe 2>/dev/null|sed 's,.*-> ,,'`"
  if [ -n "$PIDEXE" ]; then
   # Compare the inodes of $PID's executable and $SQUID to see if they're
   #  the same file.
   # Note that the syntax ${VAR%% *} removes everything in $VAR after a blank.
   # RHEL6-based systems can have a " (deleted)" as part of $PIDEXE.
   typeset PIDINODE="`ls -iL ${PIDEXE%% *} 2>/dev/null`"
   typeset SQUIDINODE="`ls -iL ${SQUIDEXE%% *}`"
   if [ "${PIDINODE%% *}" = "${SQUIDINODE%% *}" ]; then
    continue
   fi
  fi
  return 1
 done
 return 0
}

cleancache()
{
 for D in `squid_dirs $CACHE_DIR`; do
  if [ -f $D/swap.state ]; then
   # a cache exists
   echo "Checking if $NUMSQUIDS Frontier Squid${PLURAL} running..."
   if $SQUID -k parse; then
    if isrunning; then
     echo "Squid${PLURAL} running, doing nothing"
     exit 1
    fi
   else
    RETVAL=$?
    echo "ERROR parsing $CONF_FILE" >&2
    exit $RETVAL
   fi
   echo "Start clearing Cache${PLURAL} of $NUMSQUIDS Frontier Squid${PLURAL} in the background... "
   $SQUID -k shutdown 2>/dev/null||true  # in case all are not shut down
   for D in `squid_dirs $CACHE_DIR`; do
    DELD=$D/.deleting/$$
    rm -rf $DELD
    mkdir -p $DELD
    mv $D/* $DELD 2>/dev/null
   done
   nohup $SHELL -c "for D in `squid_dirs $CACHE_DIR`; do
    rm -rf \$D/.deleting
   done" </dev/null >/dev/null 2>&1 &
   echo "started"
   break
  fi
 done
 echo "Initializing Cache${PLURAL}..."
 $SQUID -z
 sleep 3
}

removecache()
{
 if isrunning; then
  echo "Squid${PLURAL} running, doing nothing"
  exit 1
 fi
 for D in `squid_dirs $CACHE_DIR`; do
  echo "Removing cache in $D..."
  rm -rf $D/* $D/.??*
 done
}

start_squid()
{
 echo "Starting $NUMSQUIDS Frontier Squid${PLURAL}... "
 $SQUID -DF
 # Allow some time for squid initialization to fail.  Also, background
 #  cache cleaning can cause the start to be quite slow and it might take
 #  extra time for the executables to start when using multiple squids.
 sleep 10
 isrunning
 RETVAL=$?
 if [ $RETVAL != 0 ] ; then
  echo "Frontier Squid start failed!!!" >&2
  $SQUID -k shutdown 2>/dev/null || true  # in case all are not shut down
 else
  if [ "$SETSQUIDAFFINITY" = true ]; then
   let AFFINITYCORE=0
   for D in `squid_dirs $PID_FILENAME_DIR`; do
    taskset -cp $AFFINITYCORE `cat $D/$(basename $PID_FILENAME)` >/dev/null
    let AFFINITYCORE+=1
   done
  fi

  sleep 2
  echo "done"
 fi

 [ $RETVAL == 0 ]
}

stop_squid()
{
 echo -n "Stopping $NUMSQUIDS Frontier Squid${PLURAL}... "
 $SQUID -k shutdown
 RETVAL=$?
 if [ $RETVAL != 0 ] ; then
  echo "Frontier Squid stop failed!!!" >&2
 else
  sleep 30 # wait for the default shutdown_lifetime
  echo "done"
 fi
 [ $RETVAL == 0 ]
}

hasawstats()
{
 [ -n "$AWSTATS_SCRIPT" ] && [ -x "$AWSTATS_SCRIPT" ] && [ "$ACCESS_LOG" != none ] && [ -z "2" ]
}

callawstats()
{
 if hasawstats; then
  for LDIR in $(squid_dirs $ACCESS_LOG_DIR); do
   typeset LFILE=$LDIR/`basename $ACCESS_LOG`
   typeset AWSTATS_CMD="${AWSTATS_SCRIPT} ${LFILE}$1"
   echo "Calling ${AWSTATS_CMD}"
   ${AWSTATS_CMD}
  done
 fi
}

rotate()
{
 # need to first make sure no other rotate is running
 typeset LOCKFILE TMPLOCK
 LOCKFILE=$ACCESS_LOG_DIR/.rotatelock
 TMPLOCK="$LOCKFILE.$$"
 trap "rm -f $TMPLOCK" 0
 echo "$$" >$TMPLOCK
 if ! ln $TMPLOCK $LOCKFILE 2>/dev/null; then
  typeset OTHERPID
  OTHERPID="`cat $LOCKFILE`"
  if [ -z "$OTHERPID" ] || ! kill -0 "$OTHERPID" 2>/dev/null; then
   rm -f $LOCKFILE
   if ! ln $TMPLOCK $LOCKFILE 2>/dev/null; then
    echo "Attempted to break lock $LOCKFILE from $OTHERPID but failed, giving up"
    exit 1
   fi
   echo "Broke lock $LOCKFILE from $OTHERPID because process was not running"
  else
   echo "Cannot create $LOCKFILE, already created by running process $OTHERPID"
   exit 1
  fi
 fi
 rm -f $TMPLOCK
 trap 0

 echo "Rotating $NUMSQUIDS Frontier Squid Logs... "
 if $SQUID_COMPRESS_LOGS; then
  typeset DELAYCOMPRESS=""
  typeset STARTRM="0"
  typeset STARTN="0"
  typeset ROTATEN="$LOGFILE_ROTATE"
  if hasawstats; then
   # when awstats script exists, don't compress the first file
   #  because it needs to parse that file
   DELAYCOMPRESS="delaycompress"
   STARTRM="1"
  else
   # in order to look like standard logrotate, start suffix numbers from 1
   #  and save one less file
   STARTN="1"
   let ROTATEN-=1
  fi
  typeset TMPF="`mktemp`"
  trap "rm -f $TMPF*" 0
  for D in `squid_dirs $ACCESS_LOG_DIR`; do
   cat >$TMPF <<!EOF!
    $D/access.log $D/cache.log $D/store.log {
     start $STARTN
     compress
     copytruncate
     daily
     $DELAYCOMPRESS
     missingok
     rotate $ROTATEN
    }
!EOF!
   ionice -n7 logrotate -f -s $TMPF.s $TMPF
   rm -f $TMPF*
   # remove non-compressed versions in case switched over
   rm -f $D/*.log.[$STARTRM-9] $D/*.log.[1-9][0-9] $D/*.log.[1-9][0-9][0-9]
  done
  trap 0
 else
  # don't compress
  typeset ROTATESUFFIX
  let ROTATESUFFIX="$LOGFILE_ROTATE - 1"
  for D in `squid_dirs $ACCESS_LOG_DIR`; do
   # delete the log files that will be kicked out first, because
   #  otherwise squid stops serving traffic while it removes them
   rm -f $D/access.log.$ROTATESUFFIX $D/cache.log.$ROTATESUFFIX $D/store.log.$ROTATESUFFIX
   # remove compressed versions in case switched over
   rm -f $D/*.log.[1-9]*.gz
  done
  $SQUID -k rotate
  RETVAL=$?
 fi
 if [ $RETVAL != 0 ]; then
  echo "failed"
 else
  echo "done"
 fi
 callawstats .0
 rm -f $LOCKFILE
 [ $RETVAL == 0 ]
}

rotateiflarge()
{
 if [ "$ACCESS_LOG" = none ]; then
  echo "No access log to rotate"
  return
 fi
 # if any access.log is too large, rotate them all
 for LDIR in $(squid_dirs $ACCESS_LOG_DIR); do
  LFILE=$LDIR/`basename $ACCESS_LOG`
  if [ -f "$LFILE" ] && [ "$(stat -c %s $LFILE)" -gt "$SQUID_MAX_ACCESS_LOG" ]; then
    rotate
    return
  fi
 done
 echo "It's not necessary to rotate the Frontier Squid Logs"
 callawstats
}

reload()
{
 echo "Reloading $NUMSQUID Frontier Squid configuration${PLURAL}"
 $SQUID -k reconfigure
}

start()
{
 if ! $SQUID_CLEAN_CACHE_ON_START || cleancache; then
  start_squid
 fi
 [ $RETVAL == 0 ]
}

stop()
{
 stop_squid
 [ $RETVAL == 0 ]
}

restart()
{
 if stop; then
  sleep 3
  start_squid   # skip cleaning the cache on a restart
  return $?
 else
  return $?
 fi
}

case "$1" in
  start)
        start
	exit $?
        ;;
  stop)
        stop
	exit $?
        ;;
  status)
        if isrunning; then
	    echo "$NUMSQUIDS Frontier Squid${PLURAL} running"
	    exit 0
	fi 
	echo "Frontier Squid${PLURAL} not running"
        exit 1
        ;;
  restart)
	restart
	exit $?
        ;;
  reload)
	reload
        exit $?
	;;
  condrestart)
	if isrunning; then
	    restart
	    exit $?
	else
	    echo "Frontier Squid not running"
  	fi
        ;;
  cleancache)
	cleancache
	exit $?
	;;
  removecache)
	removecache
	exit $?
	;;
  rotate)
	rotate
	exit $?
	;;
  rotateiflarge)
	rotateiflarge
	exit $?
	;;
  rebuildconf)
	# already done above
        ;;
  *)
        echo $"Usage: $0 {start|stop|restart|condrestart|reload|status|cleancache|removecache|rotate|rotateiflarge}"
        exit 1
esac

